package cz.cvut.fel.alg;

import java.util.List;

/**
 * Class representing a binary tree.
 *
 * @author Petr Ryšavý, Karel Horák
 */
public class Tree<T extends Tree> {

    protected int value;
    private T leftSubTree;
    private T rightSubTree;

    protected int height = 0;

    public Tree(int value) {
        this.value = value;
    }

    public Tree(int value, T leftSubTree, T rightSubTree) {
        this.value = value;
        this.leftSubTree = leftSubTree;
        this.rightSubTree = rightSubTree;
    }

    public Tree(int value, T leftSubTree, T rightSubTree, int height) {
        this(value, leftSubTree, rightSubTree);
        this.height = height;
    }

    public int getValue() {
        return value;
    }

    public T getLeftSubTree() {
        return leftSubTree;
    }

    public T getRightSubTree() {
        return rightSubTree;
    }

    public void setLeftSubTree(T leftSubTree) {
        this.leftSubTree = leftSubTree;
    }

    public void setRightSubTree(T rightSubTree) {
        this.rightSubTree = rightSubTree;
    }

    /**
     * This method finds minimal value that is stored in the tree.
     */
    public int minValue() {
        int min = (leftSubTree == null) ? Integer.MAX_VALUE : leftSubTree.minValue();
        min = Math.min(min, (rightSubTree == null) ? Integer.MAX_VALUE : rightSubTree.minValue());
        return Math.min(min, value);
    }

    /**
     * This method sums values of all nodes of the tree.
     */
    public int sumValues() {
        int sum = value;
        sum += (leftSubTree == null) ? 0 : leftSubTree.sumValues();
        sum += (rightSubTree == null) ? 0 : rightSubTree.sumValues();
        return sum;
    }

    /**
     * Computes the number of all nodes in the tree.
     */
    public int size() {
        int size = 1;
        size += (leftSubTree == null) ? 0 : leftSubTree.size();
        size += (rightSubTree == null) ? 0 : rightSubTree.size();
        return size;
    }

    /**
     * This method performs an inorder traversal of the tree and adds values of
     * visited nodes in a list based on this ordering.
     */
    public void walkInorder(List<Integer> list) {
        if(leftSubTree != null) leftSubTree.walkInorder(list);
        list.add(value);
        if(rightSubTree != null) rightSubTree.walkInorder(list);
    }

    /**
     * This method performs an preorder traversal of the tree and adds values of
     * visited nodes in a list based on this ordering.
     */
    public void walkPreorder(List<Integer> list) {
        list.add(value);
        if(leftSubTree != null) leftSubTree.walkPreorder(list);
        if(rightSubTree != null) rightSubTree.walkPreorder(list);
    }

    /**
     * This method performs an postorder traversal of the tree and adds values of
     * visited nodes in a list based on this ordering.
     */
    public void walkPostorder(List<Integer> list) {
        if(leftSubTree != null) leftSubTree.walkPostorder(list);
        if(rightSubTree != null) rightSubTree.walkPostorder(list);
        list.add(value);
    }

    /**
     * This method computes number of leaves in the tree.
     */
    public int numberOfLeaves() {
        if(leftSubTree == null && rightSubTree == null) return 1;
        else {
            return
                    ((leftSubTree == null) ? 0 : leftSubTree.numberOfLeaves()) +
                            ((rightSubTree == null) ? 0 : rightSubTree.numberOfLeaves());
        }
    }

    /**
     * This method calculates the depth of the tree. A tree with a single node
     * with no children has depth equal to 0.
     */
    public int depth() {
        return Math.max(
                (leftSubTree == null ? 0 : leftSubTree.depth() + 1),
                (rightSubTree == null ? 0 : rightSubTree.depth() + 1)
        );
    }

    /**
     * After calling this method the tree will fulfil the heap property:
     *
     * Value of any node x is smaller than the value of x's left child and also
     * smaller than the value of x's right child.
     */
    public void heapify() {
        if(leftSubTree != null) leftSubTree.heapify();
        if(rightSubTree != null) rightSubTree.heapify();

        swapDown();
    }

    /**
     * Checks whether the current tree satisfies the AVL-tree property
     */
    public boolean isAVL() {
        // TODO implement this method
        throw new UnsupportedOperationException("Method Tree::isAVL is not implemented");
    }


    protected void swapDown() {
        T cmp = leftSubTree;
        if(cmp == null || (rightSubTree != null && rightSubTree.getValue() < cmp.getValue())) cmp = rightSubTree;
        if(cmp != null && cmp.getValue() < value) {
            int tmp = value;
            value = cmp.getValue();
            cmp.value = tmp;

            cmp.swapDown();
        }
    }

    @Override
    public String toString() {
        return "Tree{" + "value=" + value + ", leftSubTree=" + leftSubTree + ", rightSubTree=" + rightSubTree + '}';
    }
}
